/*
 * Copyright (C) 1997-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#define MEDIUM_BLOB_SIZE 16777220 /*  (2**24 + 4) */
#define BLOB_SIZE 65539 /* (2**16 + 3) */

#include "server.h"
#include <mysql/mysql.h>
#include <mysql/mysql_version.h>

extern struct gengetopt_args_info conf;
extern struct misc_meta_data *mmd;

static void *mysql_ptr = NULL;


static int com_cam(int, int, char **);
static int com_create_db(int, int, char **);
static int com_cs(int, int, char **);
static int com_da(int, int, char **);
static int com_hist(int, int, char **);
static int com_info(int, int, char **);
static int com_laa(int, int, char **);
static int com_last(int, int, char **);
static int com_ls(int, int, char **);
static int com_mbox(int, int, char **);
static int com_na(int, int, char **);
static int com_pic(int, int, char **);
static int com_picass(int, int, char **);
static int com_picch(int, int, char **);
static int com_picdel(int, int, char **);
static int com_piclist(int, int, char **);
static int com_ps(int, int, char **);
static int com_sa(int, int, char **);
static int com_sl(int, int, char **);
static int com_stradd_picadd(int, int, char **);
static int com_streams(int, int, char **);
static int com_strdel(int, int, char **);
static int com_strq(int, int, char **);
static int com_summary(int, int, char **);
static int com_upd(int, int, char **);
static int com_us(int, int, char **);
static int com_verb(int, int, char **);
static int com_vrfy(int, int, char **);

static struct command cmds[] = {
	{
		.name = "cam",
		.handler = com_cam,
		.perms = DB_READ|DB_WRITE,
		.description = "copy all metadata",
		.usage = "ca source dest1 [dest2 ...]",
		.help =
                "Copy attributes and other meta data from source file\n"
                "to destination file(s). Useful for files that have  \n"
                "been renamed.                                       \n"
	},
	{
		.name = "create_db",
		.handler = com_create_db,
		.perms = DB_READ|DB_WRITE,
		.description = "create database",
		.usage = "cdb name",
		.help =
                "Create database name containing the initial columns \n"
                "for basic interoperation with server. This command  \n"
                "has to be used only once when you install this      \n"
                "software for the very first time.                   \n"
	},
	{
		.name = "clean",
		.handler = com_vrfy,
		.perms = DB_READ | DB_WRITE,
		.description = "nuke invalid entries in database",
                .usage = "clean",
                .help =
                "If \"vrfy\" shows that you have invalid entries in  \n"
                "your database, you can get rid of them with this    \n"
                "command. Always run \"upd\" and \"vrfy\" before     \n"
                "running this command. Use with caution!             \n"
	},
	{
		.name = "cs",
		.handler = com_cs,
		.perms = AFS_WRITE | DB_READ,
		.description = "change stream",
		.usage = "cs [stream]",
		.help =
                "Selects stream \"stream\" or prints current stream  \n"
                "when no stream is given.                            \n"
	},
	{
		.name = "csp",
		.handler = com_cs,
		.perms = AFS_WRITE | DB_READ,
		.description = "change stream and play",
		.usage = "csp stream",
		.help =
                "Select stream and start playing. If stream has      \n"
                "changed, skip rest of current song.                 \n"
	},
	{
		.name = "da",
		.handler = com_da,
		.perms = DB_READ | DB_WRITE,
		.description  = "drop attribute from database",
		.usage = "da name",
		.help =
		"Use with caution. All info on attribute \"name\" will \n"
		"be lost.                                            \n"
	},
	{
		.name = "hist",
		.handler = com_hist,
		.perms = DB_READ,
		.description = "print history",
		.usage = "hist",
		.help =
                "Print list of all songs together with number of     \n"
                "days since song was last played.                    \n"
	},
	{
		.name = "info",
		.handler = com_info,
		.perms = DB_READ,
		.description = "print database info",
		.usage = "info [song]",
		.help =
                "print database informations for song. Current song  \n"
                "is used if song is not given.                       \n"
	},
	{
		.name = "la", 
		.handler = com_info, 
		.perms = DB_READ, 
		.description = "list attributes",
		.usage = "la [name]",
		.help =
                "List attributes of \"name\" or of current song when \n"
                "invoked without arguments.                          \n"
	},
	{
		.name = "laa",
		.handler = com_laa,
		.perms = DB_READ,
		.description = "list available attributes",
		.usage = "laa",
		.help = "What should I say more?\n"
		
	},
	{
		.name = "last",
		.handler = com_last,
		.perms = DB_READ,
		.description = "print list of songs, order by lastplayed",
		.usage = "last [n]",
		.help =
		"The optional number n defaults to 10 if not specified.\n"
	},
	{
		.name = "ls",
		.handler = com_ls,
		.perms = DB_READ,
		.description = "list songs matching a LIKE pattern",
		.usage = "ls [pattern]",
		.help =
		"See the documentation of mysql for the definition of \n"
		"LIKE patterns.\n"
	},
	{
		.name = "mbox",
		.handler = com_mbox,
		.perms = DB_READ,
		.description = "dump song-list in mbox format to stdout",
		.usage = "mbox [pattern]",
		.help =
                "If the (optional) pattern is given, only songs      \n"
                "which match pattern are going to be included in the \n"
                "list. Otherwise, all songs are selected.  See the   \n"
                "documentation of mysql for the definition of LIKE   \n"
                "patterns.                                           \n"
	},
	{
		.name = "na",
		.handler = com_na,
		.perms = DB_READ | DB_WRITE,
		.description = "add new attribute to database",
                .usage = "na att",
                .help =
                "This adds a column named \"att\" to your mysql      \n"
                "database. \"att\" should only contain letters and   \n"
                "numbers, in paricular, '+' and '-' are not allowed. \n"
	},
	{
		.name = "ns",
		.handler = com_ps,
		.perms = AFS_WRITE | DB_READ,
		.description = "change to next stream",
		.usage = "ns",
		.help =
		"Cycle forwards through stream list.\n"
	},
	{
		.name = "pic",
		.handler = com_pic,
		.perms = DB_READ,
		.description = "get picture by name or by identifier",
                .usage = "pic [name]",
                .help =
                "Dump jpg image that is associated to given song     \n"
                "(current song if not specified) to stdout. If       \n"
                "name starts with \"#\" it is interpreted as an      \n"
                "identifier instead and the picture having that      \n"
                "identifier is dumped to stdout.                     \n"
	},
	{
		.name = "picass",
		.handler = com_picass,
		.perms = DB_READ | DB_WRITE,
		.description = "associate a picture to songs(s)",
                .usage = "picass pic_id song1 [song2...]",
                .help =
                "Associate the picture given by \"pic_id\" to all    \n"
                "given songs.                                        \n"
	},
	{
		.name = "picch",
		.handler = com_picch,
		.perms = DB_READ | DB_WRITE,
		.description = "change name of picture",
                .usage = "picch id new_name",
                .help =
                "Asign \"new_name\" to picture with identifier \"id\".\n"
	},
	{
		.name = "picdel",
		.handler = com_picdel,
		.perms = DB_READ | DB_WRITE,
		.description = "delete picture from database",
                .usage = "picdel id1 [id2...]",
                .help =
                "Delete each given picture from database.\n"
	},
	{
		.name = "piclist",
		.handler = com_piclist,
		.perms = DB_READ,
		.description = "print list of pictures",
                .usage = "piclist",
                .help =
                "Print id, name and length of each picture contained \n"
                "in the database.                                    \n"
	},
	{
		.name = "ps",
		.handler = com_ps,
		.perms = AFS_WRITE | DB_READ,
		.description = "change to previous stream",
		.usage = "ps",
		.help =
		"Cycle backwards through stream list.\n"
	},
	{
		.name = "sa",
		.handler = com_sa,
		.perms = DB_READ | DB_WRITE,
		.description = "set/unset attributes",
		.usage = "sa attr1{+|-} [attr2{+|-} ...] [name1 ...]",
		.help
                "Set (+) or unset (-) attributes \"attr1, ...\" for    \n"
                "\"name1, ...,\" or current song if no name is given.  \n"
	},
	{
		.name = "skip",
		.handler = com_sl,
		.perms = DB_READ | DB_WRITE,
		.description = "skip subsequent song(s)",
                .usage = "skip n [stream]",
                .help =
                "Skip \"n\" songs in stream \"stream\", This is      \n"
                "equivalent to calling \"sl n stream\", followed by  \n"
                "\"us name\" for each name in sl's output.           \n"
	},
	{
		.name = "sl",
		.handler = com_sl,
		.perms = DB_READ,
		.description = "print score list",
                .usage = "sl n [s]",
                .help =
                "Print list of n songs together with their scores    \n"
                "(descending order) in stream s. Use current stream  \n"
                "is s was not given.                                 \n"
	},
	{
		.name = "stradd",
		.handler = com_stradd_picadd,
		.perms = DB_READ | DB_WRITE,
		.description = "add stream",
		.usage = "stradd name",
		.help =
                "Add stream name to list of available streams.  The  \n"
                "stream definition is read from stdin and then sent  \n"
                "to the server.                                      \n"
	},
	{
		.name = "strdel",
		.handler = com_strdel,
		.perms = DB_READ | DB_WRITE,
		.description = "delete stream",
		.usage = "strdel name",
		.help =
		"Remove stream from database.\n"
	},
	{
		.name = "streams",
		.handler = com_streams,
		.perms = DB_READ,
		.description = "list streams",
		.usage = "streams",
		.help =
                "Print list of available streams. Use \"cs\" to      \n"
                "switch to any of these.                             \n"
	},
	{
		.name = "strq",
		.handler = com_strq,
		.perms = DB_READ,
		.description = "query stream definition",
		.usage = "strq name",
		.help =
		"Print definition of stream to stdout.\n"
	},
	{
		.name = "summary",
		.handler = com_summary,
		.perms = DB_READ,
		.description = "list attributes",
		.usage = "summary",
		.help
                "Print a list of attributes together with number of  \n"
                "songs having that attribute set.                    \n"
	},
	{
		.name = "upd",
		.handler = com_upd,
		.perms = DB_READ | DB_WRITE,
		.description = "update database",
                .usage = "upd",
                .help =
                "This command uses the find_cmd which was specified  \n"
                "in dbtool.conf to find your sound files. New files  \n"
                "are then added to the mysql database. Use this      \n"
                "command if you got new sound files or if you have   \n"
                "moved some files around.                            \n"
	},
	{
		.name = "us",
		.handler = com_us,
		.perms = DB_READ | DB_WRITE,
		.description = "update lastplayed time",
                .usage = "us name",
                .help =
                "Update lastplayed time without aktually playing the \n"
                "thing.                                              \n"
	},
	{
		.name = "verb",
		.handler = com_verb,
		.perms = DB_READ | DB_WRITE,
		.description = "send verbatim sql query",
                .usage = "verb cmd",
                .help =
                "Send \"cmd\" to mysql server. For expert/debugging  \n"
                "only. Note that cmd usually must be escaped.  Use   \n"
                "only if you know what you are doing!                \n"
	},
	{
		.name = "vrfy",
		.handler = com_vrfy,
		.perms = DB_READ,
		.description = "list invalid entries in database",
                .usage = "vrfy",
                .help =
                "Show what \"clean\" would delete. Should always be  \n"
                "run after updating the database with \"upd\".       \n"
	},
	{
		.name = NULL,
	}
};

static struct regex_list regex_list[] = {
	{	.regex = "IS_N_SET",
		.replacement = "(data.%s != '1')"
	}, {
		.regex = "IS_SET",
		.replacement = "(data.%s = '1')"
	}, {
		.regex = "PICID",
		.replacement = "%sdata.Pic_Id"
	}, {
		.regex = "NAME_LIKE",
		.replacement = "(dir.name like '%s')"
	}, {
		.regex = "DIR_LIKE",
		.replacement = "(dir.dir like '%s')"
	}, {
		.regex = "LASTPLAYED",
		.replacement = "%sFLOOR((UNIX_TIMESTAMP(now())"
			"-UNIX_TIMESTAMP(data.Lastplayed))/60)"
	}, {
		.regex = "NUMPLAYED",
		.replacement = "%sdata.Numplayed"
	}, {
		.regex = NULL,
	}
};

static int real_query(char *query)
{
	if (!mysql_ptr) {
		clog(ERROR, "%s: no database active\n", __func__);
		return -1;
	}
	clog(DEBUG, "%s: %s\n", __func__, query);
	if (mysql_real_query(mysql_ptr, query, strlen(query))) {
		clog(ERROR, "%s: real_query error (%s)\n", __func__,
			mysql_error(mysql_ptr));
		return -1;
	}
	return 1;
}

/*
 * Use open connection given by mysql_ptr to query server. Returns a
 * result pointer on succes and NULL on errors
 */
static struct MYSQL_RES *get_result(char *query)
{
	void *result;

	if (real_query(query) < 0)
		return NULL;
	result = mysql_store_result(mysql_ptr);
	if (!result)
		clog(ERROR, "%s: store_result error\n", __func__);
	return result;
}
/*
 * write input from fd to dynamically allocated char array,
 * but maximal max_size byte
 */
static int fd2buf(int fd, char **buf_ptr, size_t max_size)
{
	const size_t chunk_size = 1024;
	size_t size = 2048;
	char *buf = para_malloc(size * sizeof(char)), *p = buf;
	int ret;

	while ((ret = recv_bin_buffer(fd, p, chunk_size)) > 0) {
		p += ret;
		if ((p - buf) + chunk_size > size) {
			char *tmp;

			size *= 2;
			if (size > max_size) {
				ret = -1;
				goto out;
			}
			tmp = para_realloc(buf, size);
			p = (p - buf) + tmp;
			buf = tmp;
		}
	}
	if (ret < 0)
		goto out;
	*buf_ptr = buf;
	ret = p - buf;
out:
	if (ret < 0 && buf)
		free(buf);
	return ret;
}

/*
 * new attribute
 */
static int com_na(int fd, int argc, char *argv[])
{
	char *q;
	int ret;

	if (argc < 1)
		return -1;
	q = make_message("alter table data add %s char(1) "
		"not null default 0", argv[1]);
	ret = real_query(q);
	free(q);
	return ret;
}

/*
 * delete attribute
 */
static int com_da(int fd, int argc, char *argv[])
{
	char *q;
	int ret;

	if (argc < 1)
		return -1;
	q = make_message("alter table data drop %s", argv[1]);
	ret = real_query(q);
	free(q);
	return ret;
}

/* stradd/pic_add */
static int com_stradd_picadd(int fd, int argc, char *argv[])
{
	char *blob = NULL, *esc_blob = NULL, *query = NULL, *fmt;
	int ret;
	size_t size;
	
	if (argc < 1 || strlen(argv[1]) >= MAXLINE - 1 || !mysql_ptr)
		return -1;
	if (!strcmp(argv[0], "picadd")) {
		size = MEDIUM_BLOB_SIZE;
		fmt = "insert into pics (name, pic) values ('%s','%s')";
	} else {
		size = BLOB_SIZE;
		fmt = "insert into streams (name, def) values ('%s','%s')";
	}
	if (send_buffer(fd, AWAITING_DATA_MSG) < 0)
		return -1;
	if ((ret = fd2buf(fd, &blob, size)) < 0)
		return -1;
	size = ret;
	ret = -1;
	esc_blob = para_malloc(2 * size);
	mysql_real_escape_string(mysql_ptr, esc_blob, blob, size);
	query = make_message(fmt, argv[1], esc_blob);
	if (real_query(query) < 0)
		goto out;
	ret = 1;
out:
	if (esc_blob)
		free(esc_blob);
	if (blob)
		free(blob);
	if (query)
		free(query);
	return ret;
}

/*
 * print results to fd
 */
static int print_results(int fd, void *result,
		unsigned int top, unsigned int left,
		unsigned int bottom, unsigned int right)
{
	unsigned int i,j;
	MYSQL_ROW row;

	for (i = top; i <= bottom; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			return -1;
		for (j = left; j <= right; j++)
			send_va_buffer(fd, j == left? "%s" : "\t%s",
				row[j]?  row[j] : "NULL");
		send_buffer(fd, "\n");
	}
	return 0;
}

/*
 * verbatim
 */
static int com_verb(int fd, int argc, char *argv[])
{
	void *result = NULL;
	int ret;
	unsigned int num_rows, num_fields;

	if (argc < 1)
		return -1;
        result = get_result(argv[1]);
	if (!result)
		/* return success, because it's ok to have no results */
		return 1;
	num_fields = mysql_field_count(mysql_ptr);
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (num_fields && num_rows)
		ret = print_results(fd, result, 0, 0, num_rows - 1,
			num_fields - 1);
	mysql_free_result(result);
	return ret;
}

/* returns NULL on errors or if there are no atts defined yet */
static void *get_all_attributes(void)
{
	void *result = get_result("desc data");
	unsigned int num_rows;

	if (!result)
		return NULL;
	num_rows = mysql_num_rows(result);
	if (num_rows < 5) {
		mysql_free_result(result);
		return NULL;
	}
        mysql_data_seek(result, 4); /* skip Lastplayed, Numplayed... */
	return result;
}

/*
 * list all attributes
 */
static int com_laa(int fd, int argc, char *argv[])
{
	void *result;
	int ret;

	result = get_all_attributes();
	if (!result)
		return -1;
	ret = print_results(fd, result, 0, 0, mysql_num_rows(result) - 5, 0);
	mysql_free_result(result);
	return ret;
}

/*
 * history
 */
static int com_hist(int fd, int argc, char *argv[]) {
	int ret = -1;
	void *result = NULL;
	char *q;
	unsigned int num_rows;

	q = make_message("select name,to_days(now())-to_days(lastplayed) from "
		"data where %s%s lastplayed order by lastplayed",
		(argc < 1)? "" : argv[1] , (argc < 1)? "" : "='1' and" );
        result = get_result(q);
	free(q);
	if (!result)
		return ret;
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (num_rows)
		ret = print_results(fd, result, 0, 0, num_rows - 1, 1);
	mysql_free_result(result);
	return ret;
}

/*
 * get last num songs
 */
static int com_last(int fd, int argc, char *argv[])
{
	void *result = NULL;
	char *q;
	int num, ret;

	if (argc < 1)
		num = 10;
	else
		num = atoi(argv[1]);
	if (!num)
		return -1;
	q = make_message("select name from data order by lastplayed desc "
		"limit %u", num);
	result = get_result(q);
	free(q);
	if (!result)
		return -1;
	ret = print_results(fd, result, 0, 0, mysql_num_rows(result) - 1, 0);
	mysql_free_result(result);
	return ret;
}

static int com_mbox(int fd, int argc, char *argv[])
{
	void *result;
	MYSQL_ROW row;
	int ret = -1;
	unsigned int num_rows, num_fields;
	char *query = para_strdup("select concat('From foo@localhost ', "
		"date_format(Lastplayed, '%a %b %e %T %Y'), "
		"'\nReceived: from\nTo: bar\n");

	result = get_all_attributes();
	if (!result)
		goto out;
	while ((row = mysql_fetch_row(result))) {
		char *tmp;

		if (!row[0])
			goto out;
		tmp = make_message("%s X-Attribute-%s: ', %s, '\n", query,
			row[0], row[0]);
		free(query);
		query = tmp;
	}
	query = para_strcat(query,
		"From: a\n"
		"Subject: "
		"', name, '"
		"\n\n\n"
		"') from data"
	);
	if (argc >= 1) {
		char *tmp = make_message("%s where name LIKE '%s'", query,
			argv[1]);
		free(query);
		query = tmp;
	}
	mysql_free_result(result);
        result = get_result(query);
	if (!result)
		goto out;
	num_fields = mysql_field_count(mysql_ptr);
	num_rows = mysql_num_rows(result);
	if (!num_fields || !num_rows)
		goto out;
	ret = print_results(fd, result, 0, 0, num_rows - 1, num_fields - 1);
out:
	free(query);
	if (result)
		mysql_free_result(result);
	return ret;
}

static char *escape_str(char* old)
{
	char *new;

	if (!mysql_ptr)
		return NULL;
	new = para_malloc(2 * strlen(old) * sizeof(char) + 1);

	if (new)
		mysql_real_escape_string(mysql_ptr, new, old, strlen(old));
	return new;
}

static char *escaped_basename(char *name)
{
	char *esc, *bn = para_basename(name);

	if (!bn)
		return NULL;
	esc = escape_str(bn);
	free(bn);
	return esc;
}

/* get attributes by name. If verbose is not 0, get_a writes a string
 * into atts of the form 'att1="0",att2="1"', which is used in com_cam
 * for contructing a mysql update query.
 * never returns NULL in *NON VERBOSE* mode
 */
static char *get_atts(char *name, int verbose)
{
	char *atts = NULL, *buf, *ebn;
	void *result = NULL, *result2 = NULL;
	MYSQL_ROW row, row2;
	int i, ret = -1;
	unsigned int num_fields;

	result2 = get_all_attributes();
	if (!result2)
		goto out;
	if (!(ebn = escaped_basename(name)))
		goto out;
	buf = make_message("select * from data where name='%s'", ebn);
	free(ebn);
	result = get_result(buf);
	free(buf);
	if (!result)
		goto out;
	num_fields = mysql_num_fields(result);
	if (num_fields < 5)
		goto out;
        mysql_data_seek(result2, 4); /* skip Lastplayed, Numplayed... */
	row = mysql_fetch_row(result);
	if (!row)
		goto out;
	for (i = 4; i < num_fields; i++) {
		int is_set = row[i] && !strcmp(row[i], "1");
		row2 = mysql_fetch_row(result2);
		if (!row2 || !row2[0])
			goto out;
		if (atts && (verbose || is_set))
			atts = para_strcat(atts, verbose? "," : " ");
                if (is_set || verbose)
			atts = para_strcat(atts, row2[0]);
		if (verbose)
			atts = para_strcat(atts, is_set? "=\"1\"" : "=\"0\"");
        }
	ret = 1;
out:	
	if (result2)
		mysql_free_result(result2);
	if (result)
		mysql_free_result(result);
	if (!atts && !verbose)
		atts = para_strdup("(none)");
	return atts;
}

/* never returns NULL in verbose mode */
static char *get_meta(char *name, int verbose)
{
	MYSQL_ROW row;
	void *result = NULL;
	char *ebn, *q, *ret = NULL;
	const char *verbose_fmt =
		"select concat('lastplayed: ', "
		"(to_days(now()) - to_days(lastplayed)),"
		"' day(s). numplayed: ', numplayed, "
		"', pic: ', pic_id) "
		"from data where name = '%s'";
	/* is that really needed? */
	const char *fmt = "select concat('lastplayed=\\'', lastplayed, "
		"'\\', numplayed=\\'', numplayed, "
		"'\\', pic_id=\\'', pic_id, '\\'') "
		"from data where name = '%s'";

	if (!(ebn = escaped_basename(name)))
		goto out;
	q = make_message(verbose? verbose_fmt : fmt, ebn);
	free(ebn);
        result = get_result(q);
	free(q);
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	ret = strdup(row[0]);
out:
	if (result)
		mysql_free_result(result);
	if (!ret && verbose)
		ret = para_strdup("(not yet played)");
	return ret;
}

static char *get_dir(char *name)
{
	char *ret = NULL, *q, *ebn;
	void *result;
	MYSQL_ROW row;

	if (!(ebn = escaped_basename(name)))
		return NULL;
	q = make_message("select dir from dir where name = '%s'", ebn);
	free(ebn);
	result = get_result(q);
	free(q);
	if (!result)
		return NULL;
	row = mysql_fetch_row(result);
	if (row && row[0])
		ret = para_strdup(row[0]);
	mysql_free_result(result);
	return ret;
}

/* never returns NULL */
static char *get_current_stream(void)
{
	char *ret;
	MYSQL_ROW row;
	void *result = get_result("select def from streams where "
		"name = 'current_stream'");

	if (!result)
		goto err_out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto err_out;
	ret = para_strdup(row[0]);
	mysql_free_result(result);
	return ret;
err_out:
	if (result)
		mysql_free_result(result);
	return para_strdup("(none)");
}
/*
 * Read stream definition of stream streamname and construct mysql
 * query. Return NULL on errors. If streamname is NULL, use current
 * stream. If that is also NULL, use query that selects everything.
 * If filename is NULL, query will list everything, otherwise only
 * the score of given file.
 */
static char *get_query(char *streamname, char *filename)
{
	char *accept_opts = NULL, *deny_opts = NULL, *score = NULL;
	char *select_clause, *where_clause, *order, *query;
	char command[255] = ""; /* buffer for sscanf */
	void *result;
	MYSQL_ROW row;
	char *end, *tmp;

	if (!streamname)
		tmp = get_current_stream();
	else
		tmp = para_strdup(streamname);
	if (!strcmp(tmp, "(none)")) {
		free(tmp);
		return para_strdup("select data.name, dir.dir, 'no stream' "
		"from data, dir where data.name = dir.name "
		"order by data.lastplayed");
	}
	free(tmp);
	query = make_message("select def from streams where name='%s'",
		streamname);
	result = get_result(query);
	free(query);
	query = NULL;
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	end = row[0];
	while (*end) {
		int n;
		char *arg, *line = end;
		
		if (!(end = strchr(line, '\n')))
			break;
		*end = '\0';
		end++;	
		if (sscanf(line, "%200s%n", command, &n) < 1)
			continue;
		arg = line + n;
		if (!strcmp(command, "accept:")) {
			char *tmp2 = s_a_r_list(regex_list, arg);
			if (accept_opts)
				accept_opts = para_strcat(
					accept_opts, " or ");
			accept_opts = para_strcat(accept_opts, tmp2);
			continue;
		}
		if (!strcmp(command, "deny:")) {
			char *tmp2 = s_a_r_list(regex_list, arg);
			if (deny_opts)
				deny_opts = para_strcat(deny_opts, " or ");
			deny_opts = para_strcat(deny_opts, tmp2);
			continue;
		}
		if (!strcmp(command, "score:"))
			score = s_a_r_list(regex_list, arg);
	}
	if (!score) {
		score = s_a_r_list(regex_list, conf.mysql_default_score_arg);
		if (!score)
			goto out;
	}
	if (filename) {
		char *ebn = escaped_basename(filename);
		if (!ebn)
			goto out;
		select_clause = make_message("select %s from data ", score);
		free(score);
		where_clause = make_message( "where name = '%s' ", ebn);
		free(ebn);
		order = para_strdup("");
		goto write_query;
	}
	order = make_message("order by -(%s)", score);
	select_clause = para_strdup("select concat(dir.dir, '/', dir.name) "
		"from data,dir where dir.name = data.name ");
	free(score);
	if (accept_opts && deny_opts) {
		where_clause = make_message("and ((%s) and not (%s)) ",
			accept_opts, deny_opts);
		goto write_query;
	}
	if (accept_opts && !deny_opts) {
		where_clause = make_message("and (%s) ", accept_opts);
		goto write_query;
	}
	if (!accept_opts && deny_opts) {
		where_clause = make_message("and not (%s) ", deny_opts);
		goto write_query;
	}
	where_clause = para_strdup("");
write_query:
	query = make_message("%s %s %s", select_clause, where_clause, order);
	free(order);
	free(select_clause);
	free(where_clause);
out:
	if (accept_opts)
		free(accept_opts);
	if (deny_opts)
		free(deny_opts);
	if (result)
		mysql_free_result(result);
	return query;
}



/*
 * This is called from server and from some commands. Name must not be NULL
 * Never returns NULL.
 */
static char *get_dbinfo(char *name)
{
	char *meta, *atts, *info, *dir;
	char *query, *stream = get_current_stream();
	char *score = para_strdup("(no score)");
	void *result = NULL;
	MYSQL_ROW row;

	if (!name)
		return para_strdup("(none)");
	meta = get_meta(name, 1);
	atts = get_atts(name, 0);
	dir = get_dir(name);
	/* get score */
	query = get_query(stream, name);
	if (!query)
		goto write;
	result = get_result(query);
	if (!result)
		goto write;
	row = mysql_fetch_row(result);
	if (!row || ! row[0])
		goto write;
	free(score);
	score = para_strdup(row[0]);
write:
	info = make_message("dbinfo1:dir: %s\n"
		"dbinfo2:stream: %s, %s, score: %s\n"
		"dbinfo3:%s\n",
		dir? dir : "(not contained in table)",
		stream, meta, score,
		atts);
	if (query)
		free(query);
	if (dir)
		free(dir);
	free(score);
	free(meta);
	free(atts);
	free(stream);
	if (result)
		mysql_free_result(result);
	return info;
}


/* might return NULL */
static char *get_current_song(void)
{
	char *name;
	mmd_lock();
	name = para_basename(mmd->filename);
	mmd_unlock();
	return name;
}


/* print database info */
static int com_info(int fd, int argc, char *argv[])
{
	char *name = NULL, *meta = NULL, *atts = NULL, *dir = NULL;
	int ret = -1, com_la = strcmp(argv[0], "info");
	
	if (argc < 1) {
		if (!(name = get_current_song()))
			goto out;
		if ((ret = send_va_buffer(fd, "%s\n", name)) < 0)
			goto out;
	} else
		if (!(name = escaped_basename(argv[1])))
			goto out;
	meta = get_meta(name, 1);
	atts = get_atts(name, 0);
	dir = get_dir(name);
	if (com_la)
		ret = send_va_buffer(fd, "%s\n", atts);
	else
		ret = send_va_buffer(fd, "dir: %s\n" "%s\n" "attributes: %s\n",
			dir? dir : "(not contained in table)", meta, atts);
out:
	if (meta)
		free(meta);
	if (atts)
		free(atts);
	if (dir)
		free(dir);
	if (name)
		free(name);
	return ret;
}
static int change_stream(char *stream)
{
	char *query;
	int ret = -1;
	/* try to insert if it does not exist (compatibility) */
	query = make_message("insert into streams (name, def) values "
		"('current_stream', '%s')", stream);
	real_query(query); /* ignore return value */
	free(query);
	query = make_message("update streams set def='%s' "
		"where name = 'current_stream'", stream);
	ret = real_query(query);
	free(query);
	return ret;
}

static long unsigned get_pic_id_by_songname(char *name)
{
	char *q, *ebn;
	void *result = NULL;
	long unsigned ret;
	MYSQL_ROW row;

	if (!(ebn = escaped_basename(name)))
		return -1;
	q = make_message("select pic_id from data where name = '%s'", ebn);
	free(ebn);
	result = get_result(q);
	free(q);
	if (!result)
		return -1;
	row = mysql_fetch_row(result);
	ret = -1;
	if (row && row[0])
		ret = atol(row[0]);
	mysql_free_result(result);
	return ret;
}

/*
 * picass: associate pic to song
 */
static int com_picass(int fd, int argc, char *argv[])
{
	char *q, *ebn;
	long unsigned id;
	int i, ret;

	if (argc < 2)
		return -1;
	id = atol(argv[1]);
	for (i = 2; i <= argc; i++) {
		ebn = escaped_basename(argv[i]);
		if (!ebn)
			return -1;
		q = make_message("update data set pic_id = %lu "
				"where name=  '%s'", id, ebn);
		free(ebn);
		ret = real_query(q);
		free(q);
		if (ret < 0)
			return -1;
	}
	return 1;
}

/*
 * picch: change entry's name in pics table
 */
static int com_picch(int fd, int argc, char *argv[])
{
	int ret;
	long unsigned id;
	char *q;

	if (argc != 2)
		return -1;
	id = atol(argv[1]);
	if (strlen(argv[2]) > MAXLINE)
		return -1;
	ret = -1;
	q = make_message("update pics set name='%s' where id=%lu", argv[2], id);
	ret = real_query(q);
	free(q);
	return ret;
}

/*
 * piclist: print list of pics in db
 */
static int com_piclist(int fd, int argc, char *argv[])
{
	void *result = NULL;
	MYSQL_ROW row;
	unsigned long *length;
	int ret = -1;

	result = get_result("select id,name,pic from pics order by id");
	if (!result)
		return -1;
	while ((row = mysql_fetch_row(result))) {
		length = mysql_fetch_lengths(result);
		if (!row || !row[0] || !row[1] || !row[2])
			continue;
		send_va_buffer(fd, "%s\t%lu\t%s\n", row[0], length[2],
			row[1]);
	}
	ret = 1;
	mysql_free_result(result);
	return ret;
}

/*
 * picdel: delete picture from database
 */
static int com_picdel(int fd, int argc, char *argv[])
{
	char *q;
	long unsigned id;
	my_ulonglong aff;
	int i, ret = -1;

	if (argc < 1)
		return -1;
	for (i = 1; i <= argc; i++) {
		id = atol(argv[i]);
		q = make_message("delete from pics where id = %lu", id);
		ret = real_query(q);
		free(q);
		if (ret < 0)
			return -1;
		aff = mysql_affected_rows(mysql_ptr);
		if (!aff) {
			if (send_va_buffer(fd, "No such id: %lu\n", id) < 0)
				return -1;
			continue;
		}
	}
	return 1;
}
/*
 * pic: get picture by songname or by number
 */
static int com_pic(int fd, int argc, char *argv[])
{
	void *result = NULL;
	MYSQL_ROW row;
	unsigned long *length, id;
	int ret;
	char *q, *name = NULL;

	if (argc < 1)
		name = get_current_song();
	else
		name = para_basename(argv[1]);
	if (!name)
		return -1;
	if (*name == '#')
		id = atoi(name + 1);
	else
		id = get_pic_id_by_songname(name);
	free(name);
	if (!id)
		return -1;
	q = make_message("select pic from pics where id = '%lu'", id);
	result = get_result(q);
	free(q);
	if (!result)
		return -1;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	length = mysql_fetch_lengths(result);
	ret = send_bin_buffer(fd, row[0], *length);
out:
	mysql_free_result(result);
	return ret;
}

/* strdel */
static int com_strdel(int fd, int argc, char *argv[])
{
	char *tmp;
	int ret = -1;

	if (argc < 1)
		return -1;
	tmp = make_message("delete from streams where name='%s'", argv[1]);
	ret = real_query(tmp);
	free(tmp);
	if (ret < 0)
		goto out;
	tmp = get_current_stream();
	ret = 1;
	if (strcmp(tmp, "(none)") && !strcmp(tmp, argv[1]))
		ret = change_stream("(none)");
out:
	return ret;
}

/*
 * ls
 */
static int com_ls(int fd, int argc, char *argv[])
{
	char *q;
	void *result;
	int ret = -1;
	unsigned int num_rows;

	if (argc > 0)
		q = make_message("select name from data where name LIKE '%s'",
			argv[1]);
	else
		q = para_strdup("select name from data");
        result = get_result(q);
	free(q);
	if (!result)
		return -1;
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (num_rows)
		ret = print_results(fd, result, 0, 0, num_rows - 1, 0);
	mysql_free_result(result);
	return ret;
}
/*
 * summary
 */
static int com_summary(int fd, int argc, char *argv[])
{
	MYSQL_ROW row;
	MYSQL_ROW row2;
	void *result;
	void *result2 = NULL;
	const char *fmt = "select count(name) from data where %s='1'";
	int ret = -1;

	result = get_all_attributes();
	if (!result)
		goto out;
	while ((row = mysql_fetch_row(result))) {
		char *buf;

		if (!row[0])
			goto out;
		buf = make_message(fmt, row[0]);
		result2 = get_result(buf);
		free(buf);
		if (!result2)
			goto out;
		row2 = mysql_fetch_row(result2);
		if (!row2 || !row2[0])
			goto out;
		ret = send_va_buffer(fd, "%s\t%s\n", row[0], row2[0]);
		if (ret < 0)
			goto out;
	}	
	ret = 1;
out:	
	if (result2)
		mysql_free_result(result2);
	if (result)
		mysql_free_result(result);
	return ret;
}

static int get_numplayed(char *name)
{
	void *result;
	MYSQL_ROW row;
	const char *fmt = "select Numplayed from data where name = '%s'";
	char *buf = make_message(fmt, name);
	int ret = -1;

        result = get_result(buf);
	free(buf);
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	ret = atoi(row[0]);
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

static int update_song(char *name)
{
	int ret = -1;
	const char *fmt1 = "update data set lastplayed = now() where name='%s'";
	const char *fmt2 = "update data set Numplayed=%i where name='%s'";
	char *q;
	char *ebn = escaped_basename(name);

	if (!ebn)
		goto out;
	q = make_message(fmt1, ebn);
	ret = real_query(q);
	free(q);
	if (ret < 0)
		goto out;
	ret = get_numplayed(ebn);
	if (ret < 0)
		goto out;
	q = make_message(fmt2, ret + 1, ebn);
	ret = real_query(q);
	free(q);
out:
	if (ebn)
		free(ebn);
	return ret;
}
/* If called as child, mmd_lock must be held */
static void update_mmd(char *info)
{
	clog(DEBUG, "%s: updating shared memory area\n", __func__);
	strncpy(mmd->dbinfo, info, MMD_INFO_SIZE - 1);
	mmd->dbinfo[MMD_INFO_SIZE - 1] = '\0';
}

static void update_song_server_handler(char *name)
{
	char *info;
	info = get_dbinfo(name);
	update_mmd(info);
	free(info);
	update_song(name);
}

static int com_us(int fd, int argc, char *argv[])
{
	if (argc != 1)
		return -1;
	return update_song(argv[1]);
}

static void refresh_mmd_dbinfo(void)
{
	char *name = get_current_song();
	char *info;

	if (!name)
		return;
	info = get_dbinfo(name);
	free(name);
	mmd_lock();
	update_mmd(info);
	mmd_unlock();
	free(info);
}

/* select previous/next stream */
static int com_ps(int fd, int argc, char *argv[])
{
	char *query, *stream = get_current_stream();
	void *result = get_result("select name from streams");
	MYSQL_ROW row;
	int match = -1, ret = -1, i;
	unsigned int num_rows;

	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if (num_rows < 2)
		goto out;
	for (i = 0; i < num_rows; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			goto out;
		if (!strcmp(row[0], "current_stream"))
			continue;
		if (!strcmp(row[0], stream)) {
			match = i;
			break;
		}
	}
	if (match < 0)
		goto out;
	if (!strcmp(argv[0], "ps"))
		i = match > 0? match - 1 : num_rows - 1;
	else
		i = match < num_rows - 1? match + 1 : 0;
        mysql_data_seek(result, i);
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	if (!strcmp(row[0], "current_stream")) {
		if (!strcmp(argv[0], "ps")) {
			i = match - 2;
			i = i < 0? i + num_rows : i;
		} else {
			i = match + 2;
			i = i > num_rows - 1? i - num_rows : i;
		}
		mysql_data_seek(result, i);
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			goto out;
	}
	query = make_message("update streams set def='%s' where name = "
		"'current_stream'", row[0]);
	ret = real_query(query);
	free(query);
	refresh_mmd_dbinfo();
out:
	free(stream);
	if (result)
		mysql_free_result(result);
	return ret;
}

/* streams */
static int com_streams(int fd, int argc, char *argv[])
{
	unsigned int num_rows;
	int i, ret = -1;
	void *result = get_result("select name from streams");
	MYSQL_ROW row;

	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (!num_rows)
		goto out;
	ret = -1;
	for (i = 0; i < num_rows; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			goto out;
		if (strcmp(row[0], "current_stream"))
			send_va_buffer(fd, "%s\n", row[0]);
	}
	ret = 1;
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/* query stream definition */
static int com_strq(int fd, int argc, char *argv[])
{
	MYSQL_ROW row;
	char *query;
	void *result;
	int ret = -1;

	if (argc < 1)
		return -1;
	query = make_message("select def from streams where name='%s'", argv[1]);
	result = get_result(query);
	free(query);
	if (!result)
		goto out;
	row = mysql_fetch_row(result);
	if (!row || !row[0])
		goto out;
	ret = send_va_buffer(fd, "%s", row[0]); /* no '\n' needed */
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/* change stream / change stream and play */
static int com_cs(int fd, int argc, char *argv[])
{
	int ret = -1, stream_change;
	char *query;
	char *old_stream = get_current_stream();
	int csp = !strcmp(argv[0], "csp");

	if (!argc) {
		if (csp)
			goto out;
		send_va_buffer(fd, "%s\n", old_stream);
		ret = 1;
		goto out;
	}
	query = get_query(argv[1], NULL); /* test if stream is valid */
	if (!query)
		goto out;
	free(query);
	/* stream is ok */
	stream_change = strcmp(argv[1], old_stream);
	if (stream_change) {
		ret = change_stream(argv[1]);
		if (ret < 0)
			goto out;
		refresh_mmd_dbinfo();
	}
	if (csp) {
		mmd_lock();
		mmd->afs_status_flags |= AFS_PLAYING;
		if (stream_change)
			mmd->afs_status_flags |= AFS_NEXT;
		mmd_unlock();
	}
	ret = 1;
out:
	free(old_stream);
	return ret;
}

/*
 * sl/skip
 */
static int com_sl(int fd, int argc, char *argv[])
{
	void *result = NULL;
	MYSQL_ROW row;
	int ret = -1, i, skip = !strcmp(argv[0], "skip");
	char *query, *stream;
	unsigned int num_rows, num;

	if (argc < 1)
		return -1;
	num = atoi(argv[1]);
	if (!num)
		return -1;
	stream = (argc == 1)?  get_current_stream() : para_strdup(argv[2]);
	query = get_query(stream, NULL);
	free(stream);
	if (!query)
		goto out;
	result = get_result(query);
	free(query);
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if (!num_rows)
		goto out;
	for (i = 0; i < num_rows && i < num; i++) {
		row = mysql_fetch_row(result);
		if (skip) {
			send_va_buffer(fd, "Skipping %s\n", row[0]);
			update_song(row[0]);
		} else
			send_va_buffer(fd, "%s\n", row[0]? row[0]: "BUG");
	}
	ret = 1;
out:
	if (result)
		mysql_free_result(result);
	return ret;
}

/*
 * update attributes of name
 */
static int update_atts(int fd, char *name, char *atts)
{
	int ret = -1;
	char *ebn, *q, *old, *new = NULL;

	if (!mysql_ptr)
		return -1;
	ebn = escaped_basename(name);
	if (!ebn)
		return -1;
	q = make_message("update data set %s where name = '%s'", atts, ebn);
	old = get_atts(ebn, 0);
	send_va_buffer(fd, "old: %s\n", old);
	free(old);
	ret = real_query(q);
	free(q);
	if (ret < 0)
		goto out;
	new = get_atts(ebn, 0);
	send_va_buffer(fd, "new: %s\n", new);
	free(new);
	ret = 1;
out:
	free(ebn);
	return ret;
}

/*
 * set attributes
 */
static int com_sa(int fd, int argc, char *argv[])
{
	int i, ret;
	char *atts = NULL, *name;

	if (argc < 1)
		return -1;
	for (i = 1; i <= argc; i++) {
		int unset = 0;
		char *tmp, *p =argv[i];
		int len = strlen(p);
		
 		if (!len)
			continue;
		switch (p[len - 1]) {
			case '+':
				unset = 0;
				break;
			case '-':
				unset = 1;
				break;
			default:
				goto no_more_atts;
		}
		p[len - 1] = '\0';
		tmp = make_message("%s%s='%s'", atts? "," : "", p,
			unset? "0" : "1");
		atts = para_strcat(atts, tmp);
		free(tmp);
	}
no_more_atts:
	if (!atts)
		return -1;
	if (i > argc) { /* no name given, use current song */
		if (!(name = get_current_song()))
			goto out;
		ret = update_atts(fd, name, atts);
		free(name);
	} else {
		ret = 1;
		for (; argv[i] && ret >= 0; i++)
			ret = update_atts(fd, argv[i], atts);
	}
	refresh_mmd_dbinfo();
out:
	return ret;
}

/*
 * copy attributes
 */
static int com_cam(int fd, int argc, char *argv[])
{
	char *name = NULL, *meta = NULL, *atts = NULL;
	int i, ret = -1;

	if (argc < 2)
		return -1;
	if (!(name = escaped_basename(argv[1])))
		return -1;
	if (!(atts = get_atts(name, 1)))
		goto out;
	if (!(meta = get_meta(name, 0)))
		goto out;
	for (i = 2; i <= argc; i++) {
		char *ebn, *q;
		if (!(ebn = escaped_basename(argv[i])))
			goto out;
		ret = send_va_buffer(fd, "updating %s\n", ebn);
		if (ret < 0) {
			free(ebn);
			goto out;
		}
		q = make_message("update data set %s where name = '%s'",
			meta, ebn);
		if ((ret = update_atts(fd, ebn, atts)) >= 0)
			ret = real_query(q);
		free(ebn);
		free(q);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	if (name)
		free(meta);
	if (meta)
		free(meta);
	if (atts)
		free(meta);
	return ret;
}

/*
 * verify / clean
 */
static int com_vrfy(int fd, int argc, char *argv[])
{
	char *query;
	int ret = -1, vrfy_mode = strcmp(argv[0], "clean");
	void *result = NULL;
	unsigned int num_rows;
	MYSQL_ROW row;
	char *escaped_name;

	result = get_result("select data.name from data left join dir on "
		"dir.name = data.name where dir.name is NULL");
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	if (!num_rows) {
		ret = send_buffer(fd, "No invalid entries\n");
		goto out;
	}
	if (vrfy_mode) {
		send_va_buffer(fd, "found %i invalid entr%s\n", num_rows,
			num_rows == 1? "y" : "ies");
		ret = print_results(fd, result, 0, 0, num_rows - 1, 0);
		goto out;
	}
	while ((row = mysql_fetch_row(result))) {
		ret = -1;
		if (!row[0])
			goto out;
		escaped_name = escape_str(row[0]);
		if (!escaped_name)
			goto out;
		send_va_buffer(fd, "deleting %s\n", escaped_name);
		query = make_message("delete from data where name = '%s'",
			escaped_name);
		ret = real_query(query);
		free(query);
		if (ret < 0)
			goto out;
	}

out:	
	if (result)
		mysql_free_result(result);
	return ret;
}

/*
 * update database
 */
static int com_upd(int fd, int argc, char *argv[])
{

        char *tempfile, *query = NULL;
        char buf[MAXLINE];
	int ret = -1;
	int num = 0;
	FILE *pipe = NULL;
	FILE *out_fd = NULL;
	void *result = NULL;
	pid_t pid;
	unsigned int num_rows;
	MYSQL_ROW row;

	/* FIXME */
	tempfile = make_message("%s/.paraslash/mysql.tmp.%i", getenv("HOME"),
		getpid());
	clog(DEBUG, "%s: find command: %s\n", __func__, conf.mysql_find_cmd_arg);
	if (!(out_fd = fopen(tempfile, "w"))) {
		clog(ERROR, "%s: can not open temp file %s\n", __func__,
			tempfile);
		goto out;
	}
	if (!(pipe = popen_read_cmdline_pid(&pid, conf.mysql_find_cmd_arg)))
		goto out;
	while ((fgets(buf, MAXLINE - 1, pipe)) && buf[0]) {
		char *p = strrchr(buf, '/');
		clog(DEBUG, "%s: read '%s'\n", __func__, buf);
		if (!p)
			goto out;
		*p = '\t';
		if (fputs(buf, out_fd) == EOF) {
			clog(ERROR, "%s: write errror\n", __func__);
			pclose(pipe);
			fclose(out_fd);
			goto out;
		}
		num++;
	}
	pclose(pipe);
	fclose(out_fd);
	clog(DEBUG, "%s: wrote tempfile %s (%d lines)\n", __func__, tempfile, num);
	if (real_query("delete from dir") < 0)
		goto out;
	query = make_message("load data local infile '%s' into table dir "
		"fields terminated by '\t' lines terminated by '\n' "
		"(dir, name)", tempfile);
	ret = real_query(query);
	free(query);
	if (ret < 0)
		goto out;
	clog(DEBUG, "%s: replaced table dir (%i entr%s)\n", __func__, num,
			num == 1? "y" : "ies");
	result = get_result("select dir.name from dir left join data on "
		"data.name = dir.name where data.name is NULL");
	if (!result)
		goto out;
	num_rows = mysql_num_rows(result);
	ret = 1;
	if (!num_rows) {
		send_buffer(fd, "No new entries\n");
		goto out;
	}
	while ((row = mysql_fetch_row(result))) {
		ret = -1;
		if (!row[0])
			goto out;
		send_va_buffer(fd, "New entry: %s\n", row[0]);
		query = make_message("insert into data (name, pic_id) values "
			"('%s','%s')", row[0], "1");
		ret = real_query(query);
		free(query);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	free(tempfile);
	if (result)
		mysql_free_result(result);
	return ret;
}

static char **server_get_song_list(int num)
{
	char **list = para_malloc((num + 1)
		* sizeof(char *));
	char *query, *stream = get_current_stream();
	void *result = NULL;
	unsigned int num_rows;
	int i;
	MYSQL_ROW row;

	query = get_query(stream, NULL);
	if (!query)
		goto err_out;
	result = get_result(query);
	if (!result)
		goto err_out;
	num_rows = mysql_num_rows(result);
	if (!num_rows)
		goto err_out;
	for (i = 0; i < num_rows && i < num; i++) {
		row = mysql_fetch_row(result);
		if (!row || !row[0])
			goto err_out;
		list[i] = para_strdup(row[0]);
	}
	list[i] = NULL;
	goto success;
err_out:
	free(list);
	list = NULL;
success:
	free(stream);
	if (query)
		free(query);
	if (result)
		mysql_free_result(result);
	return list;
}

/*
 * connect to mysql server, return mysql pointer on success, NULL on errors
 */
static int init_mysql_server(void)
{
	const char *success_msg = "successfully connected to mysql server";
	mysql_ptr = mysql_init(NULL);

	if (!mysql_ptr) {
		clog(CRIT, "%s: mysql init error\n", __func__);
		return -1;
	}

	clog(DEBUG, "%s: connecting: %s@%s:%d\n", __func__, 
		conf.mysql_user_arg, conf.mysql_host_arg, conf.mysql_port_arg);
	if (!(mysql_ptr = mysql_real_connect(mysql_ptr,
			conf.mysql_host_arg,
			conf.mysql_user_arg,
			conf.mysql_passwd_arg,
			conf.mysql_database_arg,
			conf.mysql_port_arg, NULL, 0))) {
		clog(CRIT, "%s: connect error\n", __func__);
		return -1;
	}
	clog(INFO, "%s: %s\n", __func__, success_msg);
	sprintf(mmd->dbinfo, "dbinfo1:%s\ndbinfo2:\ndbinfo3:\n", success_msg);
	return 1;
}

/* create database */
static int com_create_db(int fd, int argc, char *argv[])
{
	char *query, *name;
	int ret = -1;

	if (argc < 1)
		name = "music";
	else {
		name = argv[1];
		if (strlen(name) > MAXLINE)
			goto out;
	}
	if (mysql_ptr) {
		clog(INFO, "%s: closing database\n", __func__);
		mysql_close(mysql_ptr);
	}
	/* dont use any database */
	conf.mysql_database_arg = NULL; /* leak? */
	if (init_mysql_server() < 0 || !mysql_ptr)
		goto out;
	query = make_message("create database %s", name);
	ret = real_query(query);
	free(query);
	if (ret < 0)
		goto out;
	/* reconnect with database just created */
	mysql_close(mysql_ptr);
	conf.mysql_database_arg = para_strdup(name);
	ret = -1;
	if (init_mysql_server() < 0 || !mysql_ptr)
		goto out;
	if (real_query("create table data (name varchar(255) binary not null "
			"primary key, "
			"lastplayed datetime not null default "
				"'1970-01-01', "
			"numplayed int not null default 0, "
			"pic_id bigint unsigned not null default 1)") < 0)
		goto out;
	if (real_query("create table dir (name varchar(255) binary not null "
			"primary key, dir varchar(255) default null)") < 0)
		goto out;
	if (real_query("create table pics ("
			"id bigint(20) unsigned not null primary key "
			"auto_increment, "
			"name varchar(255) binary not null, "
			"pic mediumblob not null)") < 0)
		goto out;
	if (real_query("create table streams ("
			"name varchar(255) binary not null primary key, "
			"def blob not null)") < 0)
		goto out;
	if (real_query("insert into streams (name, def) values "
			"('current_stream', '(none)')") < 0)
		goto out;
	ret = send_va_buffer(fd, "successfully created database %s\n", name);
out:
	return ret;
	
}

static void shutdown_connection(void)
{
	if (mysql_ptr) {
		clog(NOTICE, "%s: shutting down mysql connection\n", 
			__func__);
		mysql_close(mysql_ptr);
		mysql_ptr = NULL;
	}
}

int mysql_dbtool_init(struct dbtool *db)
{
	if (!conf.mysql_passwd_given) {
		clog(EMERG, "%s: no password specified\n", __func__);
		return -1;
	}
	if (!conf.mysql_find_cmd_given) {
		clog(EMERG, "%s: no find command specified\n", __func__);
		return -1;
	}
	db->name = "mysql";
	db->cmd_list = cmds;
	db->get_song_list = server_get_song_list;
	db->update_song = update_song_server_handler;
	db->shutdown = shutdown_connection;
	return init_mysql_server();
}
